@@ -1,7 +1,7 @@
 import cv2
 import numpy as np
 import os
-from typing import List, Tuple, Callable, Callable, NoReturn
+from typing import List, Tuple, Callable, Callable, NoReturn, Union
 import threading
 import json
 
@@ -33,6 +33,8 @@ TPointerHandler = Callable[[str, int, int, int, int, int], NoReturn]
 
 mydir = os.path.dirname(os.path.abspath(__file__))
 
+default_config_path = os.path.join(mydir, "config.json")
+
 # Constants
 CHECKER_DIMS = (5,4)
 
@@ -133,8 +135,8 @@ def make_checkerboard(width, height, res = 100, fname = None):
 
 DEFAULT_UI_DIR = os.path.join(mydir, "ui")
 
-def get_required_checkerboard():
-    return make_checkerboard(CHECKER_DIMS[0] + 1, CHECKER_DIMS[1] + 1)
+def save_required_checkerboard(fpath = None):
+    return make_checkerboard(CHECKER_DIMS[0] + 1, CHECKER_DIMS[1] + 1, 100, fpath)
 
 class UIInfo(object):
     def __init__(self, ui_name, root_dir = DEFAULT_UI_DIR):
@@ -147,17 +149,39 @@ class UIInfo(object):
         # Load images
         myui = to_color(cv2.imread(os.path.join(root_dir, f"{self.name}.png")))
         mask_raw = to_gray(cv2.imread(os.path.join(root_dir, f"{self.name}_mask.png")))
+        # Should we render different than what's printed?
         render_image_fp = os.path.join(root_dir, f"{self.name}_render.png")
         if os.path.isfile(render_image_fp):
             render_image = to_color(cv2.imread(render_image_fp))
         else:
             render_image = to_color(myui)
+
+        # Make mask by finding discrete colors in sorted order
         mask_levels = sorted(set(mask_raw.ravel()))
         mask = np.zeros(mask_raw.shape, dtype=np.int)
         for (level_ix, each_level) in enumerate(mask_levels):
             mask[mask_raw == each_level] = level_ix
+        
+        # Find the chessboard in the UI
         og_corners_found, og_corners = cv2.findChessboardCorners(myui, CHECKER_DIMS)
-        assert og_corners_found
+        assert og_corners_found, "UI image must contain chessboard"
+
+        # Save button names, if they exist
+        name_fp = os.path.join(root_dir, f"{self.name}_names.txt")
+        if os.path.isfile(name_fp):
+            with open(name_fp,'r') as name_f:
+                self.names = [
+                    stripped_line for stripped_line in 
+                    [
+                        line.strip() for line in name_f.read().splitlines()
+                    ] 
+                    if len(stripped_line) > 0
+                ]
+                assert len(self.names) >= len(mask_levels), (
+                    "Name file must name every button 0 going up")
+        else:
+            self.names = list(range(len(mask_levels)))
+
         self.raw_image = myui
         self.render_image = render_image
         self.mask = mask
@@ -295,15 +319,15 @@ class UIManager(object):
         if new_button != prev_button:
             if new_button is not None:
                 # Don't call enter when pointer is gone
-                on_enter_args = (pointer_name, *new_position, self.ui.width, self.ui.height, new_button)
+                on_enter_args = (pointer_name, *new_position, self.ui.width, self.ui.height, self.ui.names[new_button])
             if prev_button is not None:
                 # Don't call exit when pointer wasn't there before
-                on_exit_args = (pointer_name, *prev_position, self.ui.width, self.ui.height, prev_button)
+                on_exit_args = (pointer_name, *prev_position, self.ui.width, self.ui.height, self.ui.names[prev_button])
 
         # Check for on_hover
         else:
             if new_position is not None:
-                on_hover_args = (pointer_name, *new_position, self.ui.width, self.ui.height, new_button)
+                on_hover_args = (pointer_name, *new_position, self.ui.width, self.ui.height, self.ui.names[new_button])
         # Update the state dictionary
         if new_position is None:
             if prev_position is not None:
@@ -339,17 +363,77 @@ class ColorPointer(object):
         else:
             self.manager.update_pointer("finger", None)
 
+class UserSettings(object):
+    def __init__(self, 
+            webcam: Union[int, str], 
+            hflip: bool, 
+            vflip: bool,
+            skintone: TBGR8Color,
+            skintone_thresh: int):
+        self.webcam = webcam
+        self.hflip = hflip
+        self.vflip = vflip
+        self.skintone = skintone
+        self.skintone_thresh = skintone_thresh
+    def save(self, fpath: str):
+        with open(fpath, 'w') as f:
+            json.dump({
+                "webcam": self.webcam, 
+                "hflip": self.hflip, 
+                "vflip": self.vflip, 
+                "skintone": self.skintone, 
+                "skintone_thresh": self.skintone_thresh
+            }, f, indent=4, sort_keys=True)
+    @staticmethod
+    def load(fpath: str):
+        assert os.path.isfile(fpath), f"Config file {fpath} doesn't exist. Please use UserSetup to make one."
+        with open(fpath, 'r') as f:
+            dct = json.load(f)
+        # Make sure it's valid
+        assert tuple(sorted(dct.keys())) == tuple(
+            sorted([
+                "webcam", 
+                "hflip", 
+                "vflip", 
+                "skintone", 
+                "skintone_thresh"])
+            ), "Config file is invalid"
+        return UserSettings(**dct)
+    @staticmethod
+    def get_default():
+        return UserSettings(0, False, False, (44, 50, 85), 11)
+    
+    @staticmethod
+    def load_global_settings():
+        return UserSettings.load(default_config_path)
+
+    @staticmethod
+    def load_or_default(fpath: str):
+        try:
+            return UserSettings.load(fpath)
+        except:
+            return UserSettings.get_default()
+    def transform_feed(self, color_img):
+        (h,w,c) = color_img.shape
+        if self.hflip:
+            color_img = color_img[:,::-1,:]
+        if self.vflip:
+            color_img = color_img[::-1, :, :]
+        return color_img
+
+
 class ARTable(object):
-    def __init__(self, ui_name, on_key, webcam_id, finger_color, finger_thres):
+    def __init__(self, ui_name, on_key, user_settings: UserSettings):
         self.window_name = "ARTable"
-        self.webcam_id = webcam_id
+        self.webcam_id = user_settings.webcam
+        self.settings = user_settings
         self.ui = UIInfo(ui_name)
         self.detector = UIDetector(self.ui)
         self.manager = UIManager(self.ui)
-        self.finger = ColorPointer(self.manager, finger_color, finger_thres)
+        self.finger = ColorPointer(self.manager, user_settings.skintone, user_settings.skintone_thresh)
         self.on_key = on_key
         self.is_running = False
-        self.thread = threading.Thread(target=self._run, daemon=True)
+        self.thread = None
         self.stop_event = threading.Event()
 
         # Bubble this up for ease-of-access
@@ -369,6 +453,8 @@ class ARTable(object):
             ret, frame = cam.read(1)
             if not ret:
                 break
+            # Apply flips as needed
+            frame = self.settings.transform_feed(frame)
             self.detector.update_calibration(frame)
             unwarped = self.detector.unwarp_frame(frame)
             self.finger.update(unwarped)
@@ -390,10 +476,14 @@ class ARTable(object):
         cv2.destroyAllWindows()
         self.is_running = False
     def start_sync(self):
-        self.thread.run()
+        if not self.is_running:
+            self.thread = threading.Thread(target=self._run, daemon=True)
+            self.thread.run()
     def start_threaded(self):
-        self.stop_event.clear()
-        self.thread.start()
+        if not self.is_running:
+            self.thread = threading.Thread(target=self._run, daemon=True)
+            self.stop_event.clear()
+            self.thread.start()
         return self.stop_event
 
 class MouseTable(object):
@@ -447,16 +537,12 @@ def add_mouse_pointer(manager, window_name):
     cv2.namedWindow(window_name)
     cv2.setMouseCallback(window_name, on_mouse)
 
-def main():
-    ui_name = "piano"
-    finger_color = (44, 50, 85)
-    finger_thres = 11
-    webcam_id = 1
-
+def main(ui_name = 'piano'):
     def on_enter(pointer_name, x, y, w, h, button):
         print(f"Pointer {pointer_name} entered {button} @ ({x} / {w}, {y} / {h})")
     
-    table = ARTable(ui_name, None, webcam_id, finger_color, finger_thres)
+    settings = UserSettings.load_global_settings()
+    table = ARTable(ui_name, None, settings)
     table.on_enter(on_enter)
     table.start_sync()
 
